from sys import stderr
from urlparse import urlparse, urljoin
from os import unlink, close, write, chmod, makedirs
from os.path import basename, exists, isdir, join
from httplib import HTTPConnection
from tempfile import mkstemp
from zipfile import ZipFile
from hashlib import md5

from .SRTM3 import sref, quads, filename, datasource as srtm3_datasource

from osgeo import gdal

def datasource(lat, lon, source_dir):
    """
    """
    fmt = 'http://viewfinderpanos-index.herokuapp.com/index.php/%s.hgt'
    url = fmt % filename(lat, lon)
    
    #
    # Create a local filepath
    #
    s, host, path, p, q, f = urlparse(url)
    
    dem_dir = md5(url).hexdigest()[:3]
    dem_dir = join(source_dir, dem_dir)
    
    dem_path = join(dem_dir, basename(path))
    dem_none = dem_path[:-4]+'.404'
    
    #
    # Check if the file exists locally
    #
    if exists(dem_path):
        return gdal.Open(dem_path, gdal.GA_ReadOnly)

    if exists(dem_none):
        return None

    if not exists(dem_dir):
        makedirs(dem_dir)
        chmod(dem_dir, 0777)
    
    assert isdir(dem_dir)
    
    #
    # Grab a fresh remote copy
    #
    print >> stderr, 'Retrieving', url, 'in VFP.vfp_datasource().'
    
    conn = HTTPConnection(host, 80)
    conn.request('GET', path)
    resp = conn.getresponse()
    
    if resp.status == 404:
        # we're probably outside the coverage area, use SRTM3 instead
        print >> open(dem_none, 'w'), url
        return None
    
    print >> stderr, 'Found', resp.getheader('location'), 'X-Zip-Path:', resp.getheader('x-zip-path')

    assert resp.status in range(300, 399), (resp.status, resp.read())
    
    zip_location = urljoin(url, resp.getheader('location'))
    zip_filepath = resp.getheader('x-zip-path')
    
    #
    # Get the real zip archive
    #
    print >> stderr, 'Getting', zip_location

    s, host, path, p, q, f = urlparse(zip_location)

    conn = HTTPConnection(host, 80)
    conn.request('GET', path)
    resp = conn.getresponse()
    
    assert resp.status in range(200, 299), (resp.status, resp.read())
    
    try:
        #
        # Get the DEM out of the zip file
        #
        handle, zip_path = mkstemp(prefix='vfp-', suffix='.zip')
        write(handle, resp.read())
        close(handle)
        
        zipfile = ZipFile(zip_path, 'r')
        
        #
        # Write the actual DEM
        #
        print >> stderr, 'Extracting', zip_filepath, 'to', dem_path
    
        dem_file = open(dem_path, 'w')
        dem_file.write(zipfile.read(zip_filepath))
        dem_file.close()
        
        chmod(dem_path, 0666)
    
    finally:
        unlink(zip_path)

    #
    # The file better exist locally now
    #
    return gdal.Open(dem_path, gdal.GA_ReadOnly)

def datasources(minlon, minlat, maxlon, maxlat, source_dir):
    """ Retrieve a list of VFP or SRTM3 datasources overlapping the tile coordinate.
    """
    lonlats = quads(minlon, minlat, maxlon, maxlat)
    sources = [datasource(lat, lon, source_dir) for (lon, lat) in lonlats]
    return [ds for ds in sources if ds]



from math import pi, sin, cos
from os import unlink, close
from tempfile import mkstemp
from os.path import exists

from osgeo import gdal
from PIL import Image
import numpy

__all__ = 'data', 'tiles'

def read_slope_aspect(filename):
    """ Return arrays of slope and aspect data (both in radians) from a filename.
    """
    if not exists(filename):
        raise IOError('Missing file "%s"' % filename)
    
    ds = gdal.Open(str(filename))
    
    if ds is None:
        raise IOError('Unopenable file "%s"' % filename)
    
    slope = bytes2slope(ds.GetRasterBand(1).ReadAsArray())
    aspect = bytes2aspect(ds.GetRasterBand(2).ReadAsArray())
    
    return slope, aspect

def save_slope_aspect(slope, aspect, wkt, xform, fp, tmpdir):
    """ Save arrays of slope and aspect to a GeoTIFF file pointer.
    """
    w, h = slope.shape
    
    try:
        handle, filename = mkstemp(dir=tmpdir, prefix='slope-aspect-', suffix='.tif')
        close(handle)
        
        driver = gdal.GetDriverByName('GTiff')
        gtiff_options = ['COMPRESS=JPEG', 'JPEG_QUALITY=95', 'INTERLEAVE=BAND']
        ds_both = driver.Create(filename, w, h, 2, gdal.GDT_Byte, gtiff_options)
        
        ds_both.SetGeoTransform(xform)
        ds_both.SetProjection(wkt)
        
        band_slope = ds_both.GetRasterBand(1)
        band_slope.SetRasterColorInterpretation(gdal.GCI_Undefined)
        band_slope.WriteRaster(0, 0, w, h, slope2bytes(slope).tostring())
        
        band_aspect = ds_both.GetRasterBand(2)
        band_aspect.SetRasterColorInterpretation(gdal.GCI_Undefined)
        band_aspect.WriteRaster(0, 0, w, h, aspect2bytes(aspect).tostring())
        
        ds_both.FlushCache()
        ds_both = None # GDAL is lame about actually writing data until this object is out of scope
        fp.write(open(filename, 'r').read())
    
    finally:
        unlink(filename)

def shade_hills(slope, aspect):
    """ Convert slope and aspect to 0-1 grayscale with combined light sources.
    """
    diffuse = shade_hills_onelight(slope, aspect, 315.0, 30.0)
    specular = shade_hills_onelight(slope, aspect, 315.0, 85.0)
    
    # sharpen specular shading on slopes
    specular = numpy.power(specular, 4)

    # 40% diffuse and 60% specular
    shaded = .4 * diffuse + (.6 * specular)
    
    return shaded

def shade_hills_onelight(slope, aspect, azimuth, altitude):
    """ Convert slope and aspect to 0-1 grayscale with given sun position.
    """
    deg2rad = pi/180

    shaded = sin(altitude * deg2rad) * numpy.sin(slope) \
            + cos(altitude * deg2rad) * numpy.cos(slope) \
            * numpy.cos((azimuth - 90.0) * deg2rad - aspect)
    
    return shaded

def arr2img(ar):
    """ Convert Numeric.array to PIL.Image.
    """
    return Image.fromstring('L', (ar.shape[1], ar.shape[0]), ar.astype('b').tostring())

def slope2bytes(slope):
    """ Convert slope from floating point to 8-bit.
    
        Slope given in radians, from 0 for sheer face to pi/2 for flat ground.
    """
    return (0xFF * numpy.sin(slope + pi/2)).astype(numpy.uint8)

def aspect2bytes(aspect):
    """ Convert aspect from floating point to 8-bit.
    
        Aspect given in radians, counterclockwise from -pi at north around to pi.
    """
    return (0xFF * (aspect/pi + 1)/2).astype(numpy.uint8)

def bytes2slope(bytes):
    """ Convert slope from 8-bit to floating point.
    
        Slope returned in radians, from 0 for sheer face to pi/2 for flat ground.
    """
    return pi/2 - numpy.arcsin(bytes.astype(numpy.float32) / 0xFF)

def bytes2aspect(bytes):
    """ Convert aspect from 8-bit to floating point.
    
        Aspect returned in radians, counterclockwise from -pi at north around to pi.
    """
    return (2 * bytes.astype(numpy.float32)/0xFF - 1) * pi



""" Starting point for DEM retrieval utilities.

>>> region(32, -65)
'North_America'

>>> region(40, -100)
'North_America'

>>> region(51, 17)
'Eurasia'

>>> region(20, 100)
'Eurasia'

>>> region(55, -175)
'Eurasia'

>>> region(0, -160)
'Eurasia'

>>> region(-13, 96)
'Eurasia'

>>> region(-32, 159)
'Australia'

>>> region(-20, 63)
'Africa'

>>> region(39, -32)
'Africa'

>>> region(0, 0)
'Africa'

>>> region(-30, 167)
'Islands'

>>> region(-38, 77)
'Islands'

>>> region(-8, -15)
'Islands'

>>> region(-55, -35)
'Islands'

>>> region(20, -155)
'Islands'

>>> region(-45, -180)
'Islands'
"""
from sys import stderr
from math import floor, log
from os import unlink, close, write, chmod, makedirs
from os.path import basename, exists, isdir, join
from httplib import HTTPConnection
from urlparse import urlparse
from tempfile import mkstemp
from zipfile import ZipFile
from hashlib import md5

from TileStache.Geography import SphericalMercator

from osgeo import gdal, osr

ideal_zoom = 10 ## log(1200*360 / 256) / log(2) # ~10.7

osr.UseExceptions() # <-- otherwise errors will be silent and useless.

sref = osr.SpatialReference()
sref.ImportFromProj4('+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs')

def region(lat, lon):
    """ Return the SRTM3 region name of a given lat, lon.
    
        Map of regions:
        http://dds.cr.usgs.gov/srtm/version2_1/Documentation/Continent_def.gif
    """
    if -45 <= lat and lat < -25 and -180 <= lon and lon < -175:
        # southern hemisphere, near dateline
        return 'Islands'
    
    elif 15 <= lat and lat < 30 and -180 <= lon and lon < -150:
        # around hawaii
        return 'Islands'
    
    elif -60 <= lat and lat < -35 and -40 <= lon and lon < 80:
        # south atlantic ocean
        return 'Islands'
    
    elif -35 <= lat and lat < -5 and -30 <= lon and lon < -5:
        # mid-atlantic, between africa and south america
        return 'Islands'
    
    elif -60 <= lat and lat < -40 and 155 <= lon and lon < 180:
        # southern half of new zealand
        return 'Islands'
    
    elif -40 <= lat and lat < -25 and 165 <= lon and lon < 180:
        # northern half of new zealand
        return 'Islands'
    
    if 15 <= lat and lat < 61 and -170 <= lon and lon < -40:
        return 'North_America'
    
    elif -60 <= lat and lat < 15 and -95 <= lon and lon < -30:
        return 'South_America'
    
    elif -35 <= lat and lat < 35 and -30 <= lon and lon < 60:
        return 'Africa'
    
    elif -20 <= lat and lat < -15 and 60 <= lon and lon < 65:
        return 'Africa'
    
    elif 35 <= lat and lat < 40 and -35 <= lon and lon < -20:
        return 'Africa'
    
    elif -10 <= lat and lat < 61 and -15 <= lon and lon < 180:
        return 'Eurasia'
    
    elif -10 <= lat and lat < 61 and -180 <= lon and lon < -135:
        return 'Eurasia'
    
    elif -15 <= lat and lat < -10 and 95 <= lon and lon < 100:
        return 'Eurasia'
    
    elif -45 <= lat and lat < -10 and 110 <= lon and lon < 180:
        return 'Australia'
    
    raise ValueError('Unknown location: %s, %s' % (lat, lon))

def filename(lat, lon):
    """ Return an SRTM3 filename with no extension, e.g. "N37W123".
    """
    if lat < 0 and lon < 0:
        NS, EW = 'S', 'W'
    elif lat < 0:
        NS, EW = 'S', 'E'
    elif lat >= 0 and lon < 0:
        NS, EW = 'N', 'W'
    elif lat >= 0:
        NS, EW = 'N', 'E'

    return '%s%02d%s%03d' % (NS, abs(lat), EW, abs(lon))

def quads(minlon, minlat, maxlon, maxlat):
    """ Generate a list of southwest (lon, lat) for 1-degree quads of SRTM3 data.
    """
    lon = floor(minlon)
    while lon <= maxlon:
    
        lat = floor(minlat)
        while lat <= maxlat:
        
            yield lon, lat
        
            lat += 1
    
        lon += 1

def datasource(lat, lon, source_dir):
    """ Return a gdal datasource for an SRTM3 lat, lon corner.
    
        If it doesn't already exist locally in source_dir, grab a new one.
    """
    #
    # Create a URL
    #
    try:
        reg = region(lat, lon)
    except ValueError:
        # we're probably outside a known region
        return None

    fmt = 'http://dds.cr.usgs.gov/srtm/version2_1/SRTM3/%s/%s.hgt.zip'
    url = fmt % (reg, filename(lat, lon))
    
    #
    # Create a local filepath
    #
    s, host, path, p, q, f = urlparse(url)
    
    dem_dir = md5(url).hexdigest()[:3]
    dem_dir = join(source_dir, dem_dir)
    
    dem_path = join(dem_dir, basename(path)[:-4])
    dem_none = dem_path[:-4]+'.404'
    
    #
    # Check if the file exists locally
    #
    if exists(dem_path):
        return gdal.Open(dem_path, gdal.GA_ReadOnly)

    if exists(dem_none):
        return None

    if not exists(dem_dir):
        makedirs(dem_dir)
        chmod(dem_dir, 0777)
    
    assert isdir(dem_dir)
    
    #
    # Grab a fresh remote copy
    #
    print >> stderr, 'Retrieving', url, 'in DEM.SRTM3.datasource().'
    
    conn = HTTPConnection(host, 80)
    conn.request('GET', path)
    resp = conn.getresponse()
    
    if resp.status == 404:
        # we're probably outside the coverage area
        print >> open(dem_none, 'w'), url
        return None
    
    assert resp.status == 200, (resp.status, resp.read())
    
    try:
        #
        # Get the DEM out of the zip file
        #
        handle, zip_path = mkstemp(prefix='srtm3-', suffix='.zip')
        write(handle, resp.read())
        close(handle)
        
        zipfile = ZipFile(zip_path, 'r')
        
        #
        # Write the actual DEM
        #
        dem_file = open(dem_path, 'w')
        dem_file.write(zipfile.read(zipfile.namelist()[0]))
        dem_file.close()
        
        chmod(dem_path, 0666)
    
    finally:
        unlink(zip_path)

    #
    # The file better exist locally now
    #
    return gdal.Open(dem_path, gdal.GA_ReadOnly)

def datasources(minlon, minlat, maxlon, maxlat, source_dir):
    """ Retrieve a list of SRTM3 datasources overlapping the tile coordinate.
    """
    lonlats = quads(minlon, minlat, maxlon, maxlat)
    sources = [datasource(lat, lon, source_dir) for (lon, lat) in lonlats]
    return [ds for ds in sources if ds]

if __name__ == '__main__':

    import doctest
    doctest.testmod()


from math import pi, log
from tempfile import mkstemp
from os import close, write, remove
from urlparse import urljoin, urlparse
from os.path import join, exists
from urllib import urlopen

from TileStache.Geography import SphericalMercator

from PIL.Image import BILINEAR as resample
import numpy

from . import arr2img, read_slope_aspect, shade_hills

def get_slope_aspect(source_dir, coord):
    """ Retrieve slope and aspect for a coordinate tile in a source directory.
    
        Source directory can be a local path, absolute path or URL.
    """
    #
    # Find a file to work with
    #
    z, x, y = '%d' % coord.zoom, '%06d' % coord.column, '%06d' % coord.row
    tile_path = '/'.join((z, x[:3], x[3:], y[:3], y[3:])) + '.tiff'
    
    scheme, host, dir_path, p, q, f = urlparse(source_dir)
    
    if scheme in ('file', ''):
        # Local files are read directly
        return read_slope_aspect(join(dir_path, tile_path))
    
    if scheme != 'http':
        raise IOError('Unknown scheme "%s"' % scheme)

    try:
        # Remote tiles have to be downloaded first for GDAL.
        tile_href = urljoin(source_dir.rstrip('/')+'/', tile_path)
        handle, tile_path = mkstemp(prefix='hillup-tile-', suffix='.tiff')
        
        write(handle, urlopen(tile_href).read())
        close(handle)
        
        return read_slope_aspect(join(dir_path, tile_path))
    
    finally:
        # No matter what happens, keep the local filesystem clean.
        remove(tile_path)

def render_tile(source_dir, coord, min_zoom):
    """ Render a single tile.

        Looks for two-band slope+aspect TIFF files in the provided source
        directory and applies a basic hillshading algorithm before returning
        a simple grayscale PIL image with flat ground shaded to 50% gray.
        
        Check lower zoom levels for DEM files if the requested zoom
        is not immediately available, but stop checking at min_zoom.
        
        Source directory can be a local path, absolute path or URL.
    """
    original = coord.copy()
    
    if original.zoom < min_zoom:
        raise Exception('Unable to find a suitable DEM tile for tile %d/%d/%d at zoom %d or above.' % (original.zoom, original.column, original.row, min_zoom))
    
    while coord.zoom >= min_zoom:
        #
        # Basic hill shading
        #
        try:
            slope, aspect = get_slope_aspect(source_dir, coord)
        except IOError:
            # File not found, zoom out and try again.
            coord = coord.zoomBy(-1).container()
            continue
        else:
            shaded = shade_hills(slope, aspect)

        #
        # Flat ground to 50% gray exactly by way of an exponent.
        #
        flat = numpy.array([pi/2], dtype=float)
        flat = shade_hills(flat, flat)[0]
        exp = log(0.5) / log(flat)
        
        shaded = numpy.power(shaded, exp)

        #
        # Extract the desired tile out of the shaded image, if necessary.
        #
        h, w = shaded.shape
        
        if coord.zoom < original.zoom:
            ul = original.zoomTo(coord.zoom).left(coord.column).up(coord.row)
            lr = original.down().right().zoomTo(coord.zoom).left(coord.column).up(coord.row)
            
            left, top, right, bottom = map(int, (ul.column * w, ul.row * h, lr.column * w, lr.row * h))
            
            shaded = shaded[top:bottom, left:right]
        
        return arr2img(0xFF * shaded.clip(0, 1)).resize((w, h), resample)
    
    raise Exception('Unable to find a suitable DEM tile for tile %d/%d/%d at zoom %d or above.' % (original.zoom, original.column, original.row, min_zoom))

class Provider:
    """ TileStache provider for rendering hillshaded tiles.
        
        Uses render_tile() to do the actual rendering legwork.

        See http://tilestache.org/doc/#custom-providers for information
        on how the Provider object interacts with TileStache.

        Source directory can be a local path, absolute path or URL, and
        will be interpreted relative to the layer's configuration path.
    """
    def __init__(self, layer, source_dir):
        self.layer = layer
        
        source_dir = urljoin(layer.config.dirpath, source_dir)
        scheme, host, path, p, q, f = urlparse(source_dir)
        assert scheme in ('http', 'file', '')

        self.source_dir = path if (scheme == '') else '%(scheme)s://%(host)s%(path)s' % locals()
    
    def renderTile(self, width, height, srs, coord):
        """
        """
        if srs != SphericalMercator().srs:
            raise Exception('Tile projection must be spherical mercator, not "%(srs)s"' % locals())
        
        rendered = render_tile(self.source_dir, coord, 0)

        if rendered.size != (width, height):
            rendered = rendered.resize((width, height), resample)

        return rendered



""" Starting point for DEM retrieval utilities.
"""
from math import pi, sin, cos
from os import unlink, close
from itertools import product
from tempfile import mkstemp
from sys import modules

import NED10m, NED100m, NED1km, SRTM1, SRTM3, VFP, Worldwide

from ModestMaps.Core import Coordinate
from TileStache.Geography import SphericalMercator
from TileStache.Core import Layer, Metatile
from TileStache.Config import Configuration
from TileStache.Caches import Disk

from osgeo import gdal, osr
from PIL import Image
import numpy

from .. import save_slope_aspect

# used to prevent clobbering in /vsimem/, see:
# http://osgeo-org.1803224.n2.nabble.com/gdal-dev-Outputting-to-vsimem-td6221295.html
vsimem_counter = 1

#
# Set up some useful projections.
#

osr.UseExceptions() # <-- otherwise errors will be silent and useless.

webmerc_proj = SphericalMercator()
webmerc_sref = osr.SpatialReference()
webmerc_sref.ImportFromProj4(webmerc_proj.srs)

class SeedingLayer (Layer):
    """ Tilestache-compatible seeding layer for preparing tiled data.
    
        Intended for use in hillup-seed.py script for preparing a tile directory.
    """
    def __init__(self, demdir, tiledir, tmpdir, source, size):
        """
        """
        cache = Disk(tiledir, dirs='safe')
        config = Configuration(cache, '.')
        Layer.__init__(self, config, SphericalMercator(), Metatile(), tile_height=size)
        
        self.provider = Provider(self, demdir, tmpdir, source)

    def name(self):
        return '.'

class Provider:
    """ TileStache provider for generating tiles of DEM slope and aspect data.
    
        Source parameter can be "srtm-ned" (default) or "ned-only".

        See http://tilestache.org/doc/#custom-providers for information
        on how the Provider object interacts with TileStache.
    """
    def __init__(self, layer, demdir, tmpdir=None, source='srtm-ned'):
        self.tmpdir = tmpdir
        self.demdir = demdir
        self.source = source
    
    def getTypeByExtension(self, ext):
        if ext.lower() != 'tiff':
            raise Exception()
        
        return 'image/tiff', 'TIFF'
    
    def renderArea(self, width, height, srs, xmin, ymin, xmax, ymax, zoom):
        """ Return an instance of SlopeAndAspect for requested area.
        """
        assert srs == webmerc_proj.srs # <-- good enough for now
        
        if self.source == 'srtm-ned':
            providers = choose_providers_srtm(zoom)
        
        elif self.source == 'ned-only':
            providers = choose_providers_ned(zoom)

        elif self.source == 'vfp':
            providers = [(VFP, 1)]

        elif self.source == 'worldwide':
            providers = [(Worldwide, 1)]

        else:
            providers = load_func_path(self.source)(zoom)
        
        assert sum([proportion for (mod, proportion) in providers]) == 1.0
        
        #
        # Prepare information for datasets of the desired extent and projection.
        #
        
        xres = (xmax - xmin) / width
        yres = (ymin - ymax) / height

        area_wkt = webmerc_sref.ExportToWkt()
        buffered_xform = xmin - xres, xres, 0, ymax - yres, 0, yres
        
        #
        # Reproject and merge DEM datasources into destination datasets.
        #
        
        driver = gdal.GetDriverByName('GTiff')
        
        composite_ds = make_empty_datasource(width+2, height+2, buffered_xform, area_wkt, self.tmpdir)
        proportion_complete = 0.

        for (module, proportion) in providers:
        
            cs2cs = osr.CoordinateTransformation(webmerc_sref, module.sref)
            
            # get a lat/lon bbox buffered by one pixel on all sides
            minlon, minlat, z = cs2cs.TransformPoint(xmin - xres, ymin + yres)
            maxlon, maxlat, z = cs2cs.TransformPoint(xmax + xres, ymax - yres)
            
            #
            # Keep a version of the composite without the
            # current layer applied for later alpha-blending.
            #
            do_blending = bool(proportion_complete > 0 and proportion < 1)
            
            if do_blending:
                composite_without = composite_ds.ReadAsArray()
            
            ds_args = minlon, minlat, maxlon, maxlat, self.demdir
            
            for ds_dem in module.datasources(*ds_args):
            
                # estimate the raster density across source DEM and output
                dem_samples = (maxlon - minlon) / ds_dem.GetGeoTransform()[1]
                area_pixels = (xmax - xmin) / composite_ds.GetGeoTransform()[1]
                
                if dem_samples > area_pixels:
                    # cubic looks better squeezing down
                    resample = gdal.GRA_Cubic
                else:
                    # cubic spline looks better stretching out
                    resample = gdal.GRA_CubicSpline

                gdal.ReprojectImage(ds_dem, composite_ds, ds_dem.GetProjection(), composite_ds.GetProjection(), resample)
                ds_dem = None
            
            #
            # Perform alpha-blending if needed.
            #
            if do_blending:
                proportion_with = proportion / (proportion_complete + proportion)
                proportion_without = 1 - proportion_with
                
                composite_with = composite_ds.ReadAsArray() * proportion_with
                composite_with += composite_without * proportion_without

                composite_ds.GetRasterBand(1).WriteArray(composite_with, 0, 0)
            
            proportion_complete += proportion
                
        elevation = composite_ds.ReadAsArray()

        unlink(composite_ds.GetFileList()[0])
        composite_ds = None
        
        #
        # Calculate and save slope and aspect.
        #
        
        slope, aspect = calculate_slope_aspect(elevation, xres, yres)

        tile_xform = xmin, xres, 0, ymax, 0, yres
        
        return SlopeAndAspect(self.tmpdir, slope, aspect, area_wkt, tile_xform)

class SlopeAndAspect:
    """ TileStache response object with PIL-like save() and crop() methods.
    
        This object knows only how to save two-band 8-bit GeoTIFFs.
        
        See http://tilestache.org/doc/#custom-providers for information
        on how the SlopeAndAspect object interacts with TileStache.
    """
    def __init__(self, tmpdir, slope, aspect, wkt, xform):
        """ Instantiate with array of slope and aspect, and minimal geographic information.
        """
        self.tmpdir = tmpdir
        
        self.slope = slope
        self.aspect = aspect
        
        self.w, self.h = self.slope.shape

        self.wkt = wkt
        self.xform = xform
    
    def save(self, output, format):
        """ Save a two-band GeoTIFF to output file-like object.
        """
        if format != 'TIFF':
            raise Exception('File format other than TIFF for slope and aspect: "%s"' % format)
        
        save_slope_aspect(self.slope, self.aspect, self.wkt, self.xform, output, self.tmpdir)
    
    def crop(self, box):
        """ Returns a rectangular region from the current image.
        
            Box is a 4-tuple with left, upper, right, and lower pixels.
            Not yet implemented!
        """
        raise NotImplementedError()

def choose_providers_srtm(zoom):
    """ Return a list of data sources and proportions for given zoom level.
        
        Each data source is a module such as SRTM1 or SRTM3, and the proportions
        must all add up to one. Return list has either one or two items.
    """
    if zoom <= SRTM3.ideal_zoom:
        return [(SRTM3, 1)]

    elif SRTM3.ideal_zoom < zoom and zoom < SRTM1.ideal_zoom:
        #bottom, top = SRTM3, SRTM1 # SRTM1 looks terrible
        bottom, top = SRTM3, NED10m

    elif zoom == SRTM1.ideal_zoom:
        #return [(SRTM1, 1)] # SRTM1 looks terrible
        bottom, top = SRTM3, NED10m

    elif SRTM1.ideal_zoom < zoom and zoom < NED10m.ideal_zoom:
        #bottom, top = SRTM1, NED10m # SRTM1 looks terrible
        bottom, top = SRTM3, NED10m

    elif zoom >= NED10m.ideal_zoom:
        return [(NED10m, 1)]

    difference = float(top.ideal_zoom) - float(bottom.ideal_zoom)
    proportion = 1. - (zoom - float(bottom.ideal_zoom)) / difference

    return [(bottom, proportion), (top, 1 - proportion)]

def choose_providers_ned(zoom):
    """ Return a list of data sources and proportions for given zoom level.
    
        Each data source is a module such as NED10m or NED1km, and the proportions
        must all add up to one. Return list has either one or two items.
    """
    if zoom <= NED1km.ideal_zoom:
        return [(NED1km, 1)]

    elif NED1km.ideal_zoom < zoom and zoom < NED100m.ideal_zoom:
        #bottom, top = NED1km, NED100m
        bottom, top = NED1km, NED100m

    elif zoom == NED100m.ideal_zoom:
        return [(NED100m, 1)]

    elif NED100m.ideal_zoom < zoom and zoom < NED10m.ideal_zoom:
        #bottom, top = NED100m, NED10m
        bottom, top = NED100m, NED10m

    elif zoom >= NED10m.ideal_zoom:
        return [(NED10m, 1)]

    difference = float(top.ideal_zoom) - float(bottom.ideal_zoom)
    proportion = 1. - (zoom - float(bottom.ideal_zoom)) / difference

    return [(bottom, proportion), (top, 1 - proportion)]

def make_empty_datasource(width, height, xform, wkt, tmpdir):
    '''
    '''
    driver = gdal.GetDriverByName('GTiff')
    handle, filename = mkstemp(dir=tmpdir, prefix='dem-tools-hillup-data-render-', suffix='.tif')
    close(handle)

    ds = driver.Create(filename, width, height, 1, gdal.GDT_Float32)
    ds.SetGeoTransform(xform)
    ds.SetProjection(wkt)
    
    ds.GetRasterBand(1).WriteArray(numpy.ones((width, height), numpy.float32) * -9999, 0, 0)
    ds.GetRasterBand(1).SetNoDataValue(-9999)
    
    return ds

def calculate_slope_aspect(elevation, xres, yres, z=1.0):
    """ Return a pair of arrays 2 pixels smaller than the input elevation array.
    
        Slope is returned in radians, from 0 for sheer face to pi/2 for
        flat ground. Aspect is returned in radians, counterclockwise from -pi
        at north around to pi.
        
        Logic here is borrowed from hillshade.cpp:
          http://www.perrygeo.net/wordpress/?p=7
    """
    width, height = elevation.shape[0] - 2, elevation.shape[1] - 2
    
    window = [z * elevation[row:(row + height), col:(col + width)]
              for (row, col)
              in product(range(3), range(3))]
    
    x = ((window[0] + window[3] + window[3] + window[6]) \
       - (window[2] + window[5] + window[5] + window[8])) \
      / (8.0 * xres);
    
    y = ((window[6] + window[7] + window[7] + window[8]) \
       - (window[0] + window[1] + window[1] + window[2])) \
      / (8.0 * yres);

    # in radians, from 0 to pi/2
    slope = pi/2 - numpy.arctan(numpy.sqrt(x*x + y*y))
    
    # in radians counterclockwise, from -pi at north back to pi
    aspect = numpy.arctan2(x, y)
    
    return slope, aspect

def load_func_path(funcpath):
    """ Load external function based on a path.
        
        Example funcpath: "Module.Submodule:Function".
    """
    modname, objname = funcpath.split(':', 1)

    __import__(modname)
    module = modules[modname]
    _func = eval(objname, module.__dict__)
    
    if _func is None:
        raise Exception('eval(%(objname)s) in %(modname)s came up None' % locals())

    return _func



from .SRTM3 import sref, quads
from .SRTM3 import datasource as srtm3_datasource
from .VFP import datasource as vfp_datasource

def datasource(lat, lon, source_dir):
    '''
    '''
    vfp_ds = vfp_datasource(lat, lon, source_dir)
    
    if vfp_ds is not None:
        return vfp_ds
    
    return srtm3_datasource(lat, lon, source_dir)

def datasources(minlon, minlat, maxlon, maxlat, source_dir):
    """ Retrieve a list of VFP or SRTM3 datasources overlapping the tile coordinate.
    """
    lonlats = quads(minlon, minlat, maxlon, maxlat)
    sources = [datasource(lat, lon, source_dir) for (lon, lat) in lonlats]
    return [ds for ds in sources if ds]



#!/usr/bin/env python

from distutils.core import setup

version = open('VERSION').read().strip()

setup(name='Hillup',
      version=version,
      description='Retrieves and prepares digital elevation data for rendering as map tiles.',
      author='Michal Migurski',
      author_email='mike@stamen.com',
      url='https://github.com/migurski/DEM-Tools',
      requires=['ModestMaps','PIL','numpy'],
      packages=['Hillup', 'Hillup.data'],
      scripts=['hillup-seed.py'],
      download_url='https://github.com/downloads/migurski' % locals(),
      license='BSD')



#!/usr/bin/env python
"""
"""
from sys import path
from os.path import exists
from optparse import OptionParser

from TileStache import getTile
from TileStache.Geography import SphericalMercator

from ModestMaps.Core import Coordinate
from ModestMaps.Geo import Location

from Hillup.data import SeedingLayer

parser = OptionParser(usage="""%prog [options] [zoom...]

Bounding box is given as a pair of lat/lon coordinates, e.g. "37.788 -122.349
37.833 -122.246". Output is a list of tile paths as they are created.

See `%prog --help` for info.""")

defaults = dict(demdir='source', tiledir='out', tmpdir=None, source='worldwide', bbox=(37.777, -122.352, 37.839, -122.086), size=256)

parser.set_defaults(**defaults)

parser.add_option('-b', '--bbox', dest='bbox',
                  help='Bounding box in floating point geographic coordinates: south west north east, default (%.3f, %.3f, %.3f, %.3f).' % defaults['bbox'],
                  type='float', nargs=4)

parser.add_option('-d', '--dem-directory', dest='demdir',
                  help='Directory for raw source elevation files, default "%(demdir)s".' % defaults)

parser.add_option('-t', '--tile-directory', dest='tiledir',
                  help='Directory for generated slope/aspect tiles, default "%(tiledir)s". This directory will be used as the "source_dir" for Hillup.tiles:Provider shaded renderings.' % defaults)

parser.add_option('--tile-list', dest='tile_list',
                  help='Optional file of tile coordinates, a simple text list of Z/X/Y coordinates. Overrides --bbox.')

parser.add_option('-s', '--source', dest='source',
                  help='Data source for elevations. One of "srtm-ned" for SRTM and NED data, "ned-only" for US-only downsample NED, "vfp" for Viewfinder Panoramas and SRTM3, "worldwide" for combined datasets (currently SRTM3 + VFP), or a function path such as "Module.Submodule:Function". Default "%(source)s".' % defaults)

parser.add_option('--tmp-directory', dest='tmpdir',
                  help='Optional working directory for temporary files. Consider a ram disk for this.')

parser.add_option('--tile-size', dest='size', type='int',
                  help='Optional size for rendered tiles, default %(size)s.' % defaults)

def generateCoordinates(ul, lr, zooms, padding):
    """ Generate a stream of (offset, count, coordinate) tuples for seeding.
    """
    # start with a simple total of all the coordinates we will need.
    count = 0
    
    for zoom in zooms:
        ul_ = ul.zoomTo(zoom).container().left(padding).up(padding)
        lr_ = lr.zoomTo(zoom).container().right(padding).down(padding)
        
        rows = lr_.row + 1 - ul_.row
        cols = lr_.column + 1 - ul_.column
        
        count += int(rows * cols)

    # now generate the actual coordinates.
    # offset starts at zero
    offset = 0
    
    for zoom in zooms:
        ul_ = ul.zoomTo(zoom).container().left(padding).up(padding)
        lr_ = lr.zoomTo(zoom).container().right(padding).down(padding)

        for row in range(int(ul_.row), int(lr_.row + 1)):
            for column in range(int(ul_.column), int(lr_.column + 1)):
                coord = Coordinate(row, column, zoom)
                
                yield (offset, count, coord)
                
                offset += 1

if __name__ == '__main__':

    path.insert(0, '.')

    options, zooms = parser.parse_args()
    
    if options.tile_list and exists(options.tile_list):

        # read out zooms, columns, rows
        zxys = [line.strip().split('/') for line in open(options.tile_list)]
        coords = [Coordinate(*map(int, (y, x, z))) for (z, x, y) in zxys]
        tiles = [(i, len(coords), coord) for (i, coord) in enumerate(coords)]
    
    else:
        lat1, lon1, lat2, lon2 = options.bbox
        south, west = min(lat1, lat2), min(lon1, lon2)
        north, east = max(lat1, lat2), max(lon1, lon2)
    
        northwest = Location(north, west)
        southeast = Location(south, east)
        
        webmerc = SphericalMercator()
    
        ul = webmerc.locationCoordinate(northwest)
        lr = webmerc.locationCoordinate(southeast)
    
        for (i, zoom) in enumerate(zooms):
            if not zoom.isdigit():
                raise KnownUnknown('"%s" is not a valid numeric zoom level.' % zoom)
    
            zooms[i] = int(zoom)
        
        tiles = generateCoordinates(ul, lr, zooms, 0)
    
    layer = SeedingLayer(options.demdir, options.tiledir, options.tmpdir, options.source, options.size)

    for (offset, count, coord) in tiles:
        
        mimetype, content = getTile(layer, coord, 'TIFF', True)

        print coord



""" Starting point for DEM retrieval utilities.
"""
from sys import stderr
from math import floor, ceil, log
from os import unlink, close, write, makedirs, chmod
from os.path import basename, exists, isdir, join
from tempfile import mkstemp, mkdtemp
from httplib import HTTPConnection
from shutil import move, rmtree
from urlparse import urlparse
from zipfile import ZipFile
from fnmatch import fnmatch
from hashlib import md5

from TileStache.Geography import SphericalMercator

from osgeo import gdal, osr

ideal_zoom = 15 ### log(3 * 3600*360 / 256) / log(2) # ~13.9

osr.UseExceptions() # <-- otherwise errors will be silent and useless.

sref = osr.SpatialReference()
sref.ImportFromProj4('+proj=longlat +ellps=GRS80 +datum=NAD83 +no_defs')

def quads(minlon, minlat, maxlon, maxlat):
    """ Generate a list of northwest (lon, lat) for 1-degree quads of NED 10m data.
    """
    lon = floor(minlon)
    while lon <= maxlon:
    
        lat = ceil(maxlat)
        while lat >= minlat:
        
            yield lon, lat
        
            lat -= 1
    
        lon += 1

def datasource(lat, lon, source_dir):
    """ Return a gdal datasource for a NED 10m lat, lon corner.
    
        If it doesn't already exist locally in source_dir, grab a new one.
    """
    #
    # Create a URL - tdds.cr.usgs.gov looks to be a redirect from
    # http://gisdata.usgs.gov/TDDS/DownloadFile.php?TYPE=ned3f_zip&FNAME=nxxwxx.zip
    #
    # FIXME for southern/western hemispheres
    fmt = 'http://tdds.cr.usgs.gov/ned/13arcsec/float/float_zips/n%02dw%03d.zip'
    url = fmt % (abs(lat), abs(lon))
    
    #
    # Create a local filepath
    #
    s, host, path, p, q, f = urlparse(url)
    
    local_dir = md5(url).hexdigest()[:3]
    local_dir = join(source_dir, local_dir)
    
    local_base = join(local_dir, basename(path)[:-4])
    local_path = local_base + '.flt'
    local_none = local_base + '.404'
    
    #
    # Check if the file exists locally
    #
    if exists(local_path):
        return gdal.Open(local_path, gdal.GA_ReadOnly)

    if exists(local_none):
        return None

    if not exists(local_dir):
        makedirs(local_dir)
        chmod(local_dir, 0777)
    
    assert isdir(local_dir)
    
    #
    # Grab a fresh remote copy
    #
    print >> stderr, 'Retrieving', url, 'in DEM.NED10m.datasource().'
    
    conn = HTTPConnection(host, 80)
    conn.request('GET', path)
    resp = conn.getresponse()
    
    if resp.status == 404:
        # we're probably outside the coverage area
        print >> open(local_none, 'w'), url
        return None
    
    assert resp.status == 200, (resp.status, resp.read())
    
    try:
        dirpath = mkdtemp(prefix='ned10m-')
        zippath = join(dirpath, 'dem.zip')

        zipfile = open(zippath, 'w')
        zipfile.write(resp.read())
        zipfile.close()

        zipfile = ZipFile(zippath)
        
        for name in zipfile.namelist():
            if fnmatch(name, '*/*/float*.???') and name[-4:] in ('.hdr', '.flt', '.prj'):
                local_file = local_base + name[-4:]

            elif fnmatch(name, '*/float*_13.???') and name[-4:] in ('.hdr', '.flt', '.prj'):
                local_file = local_base + name[-4:]

            elif fnmatch(name, '*/float*_13'):
                local_file = local_base + '.flt'

            else:
                # don't recognize the contents of this zip file
                continue
            
            zipfile.extract(name, dirpath)
            move(join(dirpath, name), local_file)
            
            if local_file.endswith('.hdr'):
                # GDAL needs some extra hints to understand the raw float data
                hdr_file = open(local_file, 'a')
                print >> hdr_file, 'nbits 32'
                print >> hdr_file, 'pixeltype float'
        
        #
        # The file better exist locally now
        #
        return gdal.Open(local_path, gdal.GA_ReadOnly)
    
    finally:
        rmtree(dirpath)

def datasources(minlon, minlat, maxlon, maxlat, source_dir):
    """ Retrieve a list of SRTM1 datasources overlapping the tile coordinate.
    """
    lonlats = quads(minlon, minlat, maxlon, maxlat)
    sources = [datasource(lat, lon, source_dir) for (lon, lat) in lonlats]
    return [ds for ds in sources if ds]
